diff --git a/src/backend/replication/logical/relation.c b/src/backend/replication/logical/relation.c
index 46e515e..99c1bba 100644
--- a/src/backend/replication/logical/relation.c
+++ b/src/backend/replication/logical/relation.c
@@ -382,7 +382,6 @@ static void
 logicalrep_typmap_invalidate_cb(Datum arg, int cacheid, uint32 hashvalue)
 {
 	HASH_SEQ_STATUS status;
-	LogicalRepTyp *entry;
 
 	/* Just to be sure. */
 	if (LogicalRepTypMap == NULL)
@@ -390,9 +389,6 @@ logicalrep_typmap_invalidate_cb(Datum arg, int cacheid, uint32 hashvalue)
 
 	/* invalidate all cache entries */
 	hash_seq_init(&status, LogicalRepTypMap);
-
-	while ((entry = (LogicalRepTyp *) hash_seq_search(&status)) != NULL)
-		entry->typoid = InvalidOid;
 }
 
 /*
@@ -403,8 +399,6 @@ logicalrep_typmap_free_entry(LogicalRepTyp *entry)
 {
 	pfree(entry->nspname);
 	pfree(entry->typname);
-
-	entry->typoid = InvalidOid;
 }
 
 /*
@@ -435,18 +429,16 @@ logicalrep_typmap_update(LogicalRepTyp *remotetyp)
 	entry->nspname = pstrdup(remotetyp->nspname);
 	entry->typname = pstrdup(remotetyp->typname);
 	MemoryContextSwitchTo(oldctx);
-	entry->typoid = InvalidOid;
 }
 
 /*
- * Fetch type info from the cache.
+ * Fetch type name from the cache by remote type OID.
  */
-Oid
-logicalrep_typmap_getid(Oid remoteid)
+char *
+logicalrep_typmap_gettypname(Oid remoteid)
 {
 	LogicalRepTyp *entry;
 	bool		found;
-	Oid			nspoid;
 
 	/* Internal types are mapped directly. */
 	if (remoteid < FirstNormalObjectId)
@@ -455,7 +447,7 @@ logicalrep_typmap_getid(Oid remoteid)
 			ereport(ERROR,
 					(errmsg("built-in type %u not found", remoteid),
 					 errhint("This can be caused by having a publisher with a higher PostgreSQL major version than the subscriber.")));
-		return remoteid;
+		return format_type_be(remoteid);
 	}
 
 	if (LogicalRepTypMap == NULL)
@@ -469,24 +461,7 @@ logicalrep_typmap_getid(Oid remoteid)
 		elog(ERROR, "no type map entry for remote type %u",
 			 remoteid);
 
-	/* Found and mapped, return the oid. */
-	if (OidIsValid(entry->typoid))
-		return entry->typoid;
-
-	/* Otherwise, try to map to local type. */
-	nspoid = LookupExplicitNamespace(entry->nspname, true);
-	if (OidIsValid(nspoid))
-		entry->typoid = GetSysCacheOid2(TYPENAMENSP,
-										PointerGetDatum(entry->typname),
-										ObjectIdGetDatum(nspoid));
-	else
-		entry->typoid = InvalidOid;
-
-	if (!OidIsValid(entry->typoid))
-		ereport(ERROR,
-				(errcode(ERRCODE_OBJECT_NOT_IN_PREREQUISITE_STATE),
-				 errmsg("data type \"%s.%s\" required for logical replication does not exist",
-						entry->nspname, entry->typname)));
+	Assert(OidIsValid(entry->remoteid));
 
-	return entry->typoid;
+	return entry->typname;
 }
diff --git a/src/backend/replication/logical/worker.c b/src/backend/replication/logical/worker.c
index e46a62e..cd7576f 100644
--- a/src/backend/replication/logical/worker.c
+++ b/src/backend/replication/logical/worker.c
@@ -100,8 +100,8 @@ static dlist_head lsn_mapping = DLIST_STATIC_INIT(lsn_mapping);
 
 typedef struct SlotErrCallbackArg
 {
-	LogicalRepRelation *rel;
-	int			attnum;
+	LogicalRepRelMapEntry *rel;
+	int			local_attnum;
 } SlotErrCallbackArg;
 
 static MemoryContext ApplyMessageContext = NULL;
@@ -281,20 +281,38 @@ slot_fill_defaults(LogicalRepRelMapEntry *rel, EState *estate,
 static void
 slot_store_error_callback(void *arg)
 {
-	SlotErrCallbackArg *errarg = (SlotErrCallbackArg *) arg;
+	SlotErrCallbackArg	*errarg = (SlotErrCallbackArg *) arg;
+	LogicalRepRelMapEntry	*rel;
+	char			*remotetypname;
+	int			remote_attnum;
 	Oid			remotetypoid,
 				localtypoid;
 
-	if (errarg->attnum < 0)
+	/* Check case of slot_store_error_callback() is called before 
+	 * errarg.local_attnum is set. */
+	if (errarg->local_attnum <0)
+		return;
+
+	rel = errarg->rel;
+	remote_attnum = rel->attrmap[errarg->local_attnum];
+
+	/* Check if invalid attribute here, e.g. dropped column. */
+	if (remote_attnum < 0)
 		return;
 
-	remotetypoid = errarg->rel->atttyps[errarg->attnum];
-	localtypoid = logicalrep_typmap_getid(remotetypoid);
+	remotetypoid = rel->remoterel.atttyps[remote_attnum];
+
+	/* Fetch remote type name from the LogicalRepTypMap cache */
+	remotetypname = logicalrep_typmap_gettypname(remotetypoid);
+
+	/* Fetch local type OID from the local sys cache */
+	localtypoid = get_atttype(errarg->rel->localreloid, errarg->local_attnum + 1);
+
 	errcontext("processing remote data for replication target relation \"%s.%s\" column \"%s\", "
-			   "remote type %s, local type %s",
-			   errarg->rel->nspname, errarg->rel->relname,
-			   errarg->rel->attnames[errarg->attnum],
-			   format_type_be(remotetypoid),
+			   "remote type \"%s\", local type \"%s\"",
+			   rel->remoterel.nspname, rel->remoterel.relname,
+			   rel->remoterel.attnames[remote_attnum],
+			   remotetypname,
 			   format_type_be(localtypoid));
 }
 
@@ -315,8 +333,8 @@ slot_store_cstrings(TupleTableSlot *slot, LogicalRepRelMapEntry *rel,
 	ExecClearTuple(slot);
 
 	/* Push callback + info on the error context stack */
-	errarg.rel = &rel->remoterel;
-	errarg.attnum = -1;
+	errarg.rel = rel;
+	errarg.local_attnum = -1;
 	errcallback.callback = slot_store_error_callback;
 	errcallback.arg = (void *) &errarg;
 	errcallback.previous = error_context_stack;
@@ -333,9 +351,7 @@ slot_store_cstrings(TupleTableSlot *slot, LogicalRepRelMapEntry *rel,
 		{
 			Oid			typinput;
 			Oid			typioparam;
-
-			errarg.attnum = remoteattnum;
-
+			errarg.local_attnum = i;
 			getTypeInputInfo(att->atttypid, &typinput, &typioparam);
 			slot->tts_values[i] = OidInputFunctionCall(typinput,
 													   values[remoteattnum],
@@ -380,8 +396,8 @@ slot_modify_cstrings(TupleTableSlot *slot, LogicalRepRelMapEntry *rel,
 	ExecClearTuple(slot);
 
 	/* Push callback + info on the error context stack */
-	errarg.rel = &rel->remoterel;
-	errarg.attnum = -1;
+	errarg.rel = rel;
+	errarg.local_attnum = -1;
 	errcallback.callback = slot_store_error_callback;
 	errcallback.arg = (void *) &errarg;
 	errcallback.previous = error_context_stack;
@@ -404,7 +420,7 @@ slot_modify_cstrings(TupleTableSlot *slot, LogicalRepRelMapEntry *rel,
 			Oid			typinput;
 			Oid			typioparam;
 
-			errarg.attnum = remoteattnum;
+			errarg.local_attnum = i;
 
 			getTypeInputInfo(att->atttypid, &typinput, &typioparam);
 			slot->tts_values[i] = OidInputFunctionCall(typinput,
diff --git a/src/include/replication/logicalproto.h b/src/include/replication/logicalproto.h
index a9736e1..90ca358 100644
--- a/src/include/replication/logicalproto.h
+++ b/src/include/replication/logicalproto.h
@@ -55,10 +55,9 @@ typedef struct LogicalRepRelation
 /* Type mapping info */
 typedef struct LogicalRepTyp
 {
-	Oid			remoteid;		/* unique id of the type */
-	char	   *nspname;		/* schema name */
-	char	   *typname;		/* name of the type */
-	Oid			typoid;			/* local type Oid */
+	Oid	remoteid;		/* unique id of the remote type */
+	char	*nspname;		/* schema name of remote type */
+	char	*typname;		/* name of the remote type */
 } LogicalRepTyp;
 
 /* Transaction info */
diff --git a/src/include/replication/logicalrelation.h b/src/include/replication/logicalrelation.h
index 8352705..563bb4f 100644
--- a/src/include/replication/logicalrelation.h
+++ b/src/include/replication/logicalrelation.h
@@ -37,6 +37,6 @@ extern void logicalrep_rel_close(LogicalRepRelMapEntry *rel,
 					 LOCKMODE lockmode);
 
 extern void logicalrep_typmap_update(LogicalRepTyp *remotetyp);
-extern Oid	logicalrep_typmap_getid(Oid remoteid);
+extern char	*logicalrep_typmap_gettypname(Oid remoteid);
 
 #endif							/* LOGICALRELATION_H */
